type value_type =
	VaNull | VaList | VaInt8 | VaInt16 | VaInt32 | VaExtended |
	VaString | VaIdent | VaFalse | VaTrue | VaBinary | VaSet | VaLString |
	VaNil | VaCollection | VaSingle | VaCurrency | VaDate | VaWString |
	VaInt64 | VaUTF8String | VaDouble;;

let min_value = VaNull;;
let max_value = VaDouble;;

type filer_flag = FfInherited | FfChildPos | FfInline;;
type filer_flags = int;; (* set of filer_flag *)

let set_of_ff (x: filer_flag): filer_flags = (
	1 lsl (Obj.magic x)
);;

let filer_signature = "TPF0";;

type byte_array = (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout)
	Bigarray.Array1.t;;

type variant = [
	| `bool of bool
	| `binary of byte_array
	| `char of char
	| `currency of float
	| `efloat of float * (int64 * int)
	| `float of float
	| `int of int
	| `int32 of int32
	| `int64 of int64
	| `nil
	| `null
	| `sfloat of float
	| `string of string
	| `tag of string
	| `time of float
	| `wchar of int
	| `wstring of Unicode.utf16_string];;

let datetime_of_unixtime (x: float): float = (
	x /. 86400.0 +. 25569.0
);;

let unixtime_of_datetime (x: float): float = (
	(x -. 25569.0) *. 86400.0
);;

let extended_of_float (x: float): int64 * int = (
	let i = Int64.bits_of_float x in
	if i = 0L then ( (* +0.0 *)
		0L, 0
	) else if i = 0x8000000000000000L then ( (* -0.0 *)
		0L, 0x8000
	) else (
		let m = Int64.logand 0x000fffffffffffffL i in
		let e = Int64.logand 0x7ff0000000000000L i in
		let s = Int64.logand 0x8000000000000000L i in
		Int64.logor 0x8000000000000000L (Int64.shift_left m 11),
		let de = Int64.add 0x3c00L (Int64.shift_right_logical e 52) in
		let ds = Int64.shift_right_logical s 48 in
		Int64.to_int (Int64.logor ds de)
	)
);;

let float_of_extended (m, e: int64 * int): float = (
	let dm = Int64.shift_right_logical (Int64.logand m 0x7fffffffffffffffL) 11 in
	let de = (
		if m = 0L then 0L else
		Int64.shift_left (Int64.sub (Int64.of_int e) 0x3c00L) 52
	) in
	let ds = Int64.shift_left (Int64.of_int (0x8000 land e)) 48 in
	Int64.float_of_bits (Int64.logor (Int64.logor ds de) dm)
);;

module Writer = struct
	
	type write_handler = string -> int -> int -> unit;;
	
	let write_raw_byte (temp: string) (output: write_handler) (x: int): unit = (
		temp.[0] <- char_of_int (x land 0xff);
		output temp 0 1
	);;
	
	let write_raw_int8 = write_raw_byte (String.create 1);;
	
	let write_raw_int16 (output: write_handler) (x: int): unit = (
		let s = String.create 2 in
		s.[0] <- char_of_int (0xff land x);
		s.[1] <- char_of_int (0xff land (x lsr 8));
		output s 0 2
	);;
	
	let write_raw_int32 (output: write_handler) (x: int32): unit = (
		let s = String.create 4 in
		s.[0] <- char_of_int (Int32.to_int (Int32.logand 0xffl x));
		s.[1] <- char_of_int (Int32.to_int (Int32.logand 0xffl (Int32.shift_right x 8)));
		s.[2] <- char_of_int (Int32.to_int (Int32.logand 0xffl (Int32.shift_right x 16)));
		s.[3] <- char_of_int (Int32.to_int (Int32.logand 0xffl (Int32.shift_right x 24)));
		output s 0 4
	);;
	
	let write_raw_int64 (output: write_handler) (x: int64): unit = (
		let s = String.create 8 in
		s.[0] <- char_of_int (Int64.to_int (Int64.logand 0xffL x));
		s.[1] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 8)));
		s.[2] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 16)));
		s.[3] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 24)));
		s.[4] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 32)));
		s.[5] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 40)));
		s.[6] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 48)));
		s.[7] <- char_of_int (Int64.to_int (Int64.logand 0xffL (Int64.shift_right x 56)));
		output s 0 8
	);;
	
	let write_raw_extended_image (output: write_handler) (m, e: int64 * int) = (
		write_raw_int64 output m;
		write_raw_int16 output e
	);;
	
	let write_raw_extended (output: write_handler) (x: float) = (
		write_raw_extended_image output (extended_of_float x)
	);;
	
	let write_raw_short_string (output: write_handler) (x: string): unit = (
		let length = String.length x in
		assert (length <= 255);
		write_raw_byte (String.create 1) output length;
		output x 0 length
	);;
	
	let write_raw_zero_terminated_string (output: write_handler)
		(x: string): unit =
	(
		output x 0 (String.length x);
		output "\x00" 0 1
	);;

	let write_resource_header (output: write_handler)
		(name: string) (size: int): unit =
	(
		output "\xff\x0a\x00" 0 3;
		write_raw_zero_terminated_string output name;
		output "\x30\x10" 0 2;
		write_raw_int32 output (Int32.of_int size)
	);;

	let write_value (output: write_handler) (x: value_type): unit = (
		write_raw_byte (String.create 1) output (Obj.magic x)
	);;

	let write_int8 (output: write_handler) (x: int): unit = (
		write_value output VaInt8;
		write_raw_int8 output x
	);;

	let write_int16 (output: write_handler) (x: int): unit = (
		if x >= -0x80 && x <= 0x7f then (
			write_int8 output x
		) else (
			write_value output VaInt16;
			write_raw_int16 output x
		)
	);;

	let write_int32 (output: write_handler) (x: int32): unit = (
		if x >= -0x8000l && x <= 0x7fffl then (
			write_int16 output (Int32.to_int x)
		) else (
			write_value output VaInt32;
			write_raw_int32 output x
		)
	);;

	let write_int64 (output: write_handler) (x: int64): unit = (
		if x >= -0x80000000L && x < 0x7f000000L then (
			write_int32 output (Int64.to_int32 x)
		) else (
			write_value output VaInt64;
			write_raw_int64 output x
		)
	);;

	let write_int (output: write_handler) (x: int): unit = (
		if x >= -0x8000 && x <= 0x7fff then (
			write_int16 output x
		) else (
			write_int32 output (Int32.of_int x)
		)
	);;

	let write_string (output: write_handler) (x: string): unit = (
		let length = String.length x in
		let non_ascii = (
			let rec loop i = (
				if i >= length then true else
				if x.[i] >= '\x80' then false else
				loop (i + 1)
			) in
			loop 0
		) in
		if not non_ascii then (
			write_value output VaUTF8String;
			write_raw_int32 output (Int32.of_int length);
			output x 0 length
		) else if length <= 255 then (
			write_value output VaString;
			write_raw_short_string output x
		) else (
			write_value output VaLString;
			write_raw_int32 output (Int32.of_int length);
			output x 0 length
		)
	);;

	let write_wide_string (output: write_handler)
		(x: Unicode.utf16_string): unit =
	(
		let utf8 = Unicode.utf8_of_utf16 x in
		let len = Unicode.UTF16.length x in
		let byte_len = 2 * len in
		if String.length utf8 < byte_len then (
			write_string output utf8
		) else (
			write_value output VaWString;
			write_raw_int32 output (Int32.of_int len);
			let data = String.create byte_len in
			for i = 0 to len - 1 do
				let c = Unicode.UTF16.get x i in
				data.[i * 2] <- char_of_int (c land 0xff);
				data.[i * 2 + 1] <- char_of_int (c lsr 8)
			done;
			output data 0 byte_len
		)
	);;

	let write_char (output: write_handler) (x: char): unit = (
		write_string output (String.make 1 x)
	);;

	let write_wchar (output: write_handler) (x: int): unit = (
		let s = Unicode.UTF16.create 2 in
		let len = ref 0 in
		Unicode.UTF16.set_code s len x;
		write_wide_string output (Unicode.UTF16.sub s 0 (!len - 1))
	);;
	
	let write_bool (output: write_handler) (x: bool): unit = (
		write_value output (if x then VaTrue else VaFalse)
	);;
	
	let write_ident (output: write_handler) (x: string): unit = (
		let u = String.uppercase x in
		if u = "NIL" then write_value output VaNil else
		if u = "NULL" then write_value output VaNull else
		if u = "FALSE" then write_value output VaFalse else
		if u = "TRUE" then write_value output VaTrue else (
			write_value output VaIdent;
			write_raw_short_string output x
		)
	);;

	let write_single (output: write_handler) (x: float): unit = (
		write_value output VaSingle;
		write_raw_int32 output (Int32.bits_of_float x)
	);;

	let write_double (output: write_handler) (x: float): unit = (
		write_value output VaDouble;
		write_raw_int64 output (Int64.bits_of_float x)
	);;
	
	let write_extended_image (output: write_handler) (x: int64 * int): unit = (
		write_value output VaExtended;
		write_raw_extended_image output x
	);;
	
	let write_extended (output: write_handler) (x: float): unit = (
		write_value output VaExtended;
		write_raw_extended output x
	);;
	
	let write_currency (output: write_handler) (x: float): unit = (
		write_value output VaCurrency;
		write_raw_int64 output (Int64.of_float (x *. 10000.0))
	);;
	
	let write_date (output: write_handler) (x: float): unit = (
		write_value output VaDate;
		write_raw_int64 output (Int64.bits_of_float (datetime_of_unixtime x))
	);;
	
	let write_binary (output: write_handler) (x: byte_array): unit = (
		write_value output VaBinary;
		let length = Bigarray.Array1.dim x in
		write_raw_int32 output (Int32.of_int length);
		let temp = String.create 1 in
		for i = 0 to length - 1 do
			write_raw_byte temp output (Bigarray.Array1.get x i)
		done
	);;
	
	let write_nil (output: write_handler): unit = (
		write_value output VaNil
	);;
	
	let write_null (output: write_handler): unit = (
		write_value output VaNull
	);;
	
	let write_property_name = write_raw_short_string;;
	
	let write_list_begin (output: write_handler): unit = (
		write_value output VaList
	);;
	
	let write_list_end (output: write_handler): unit = (
		output "\x00" 0 1
	);;
	
	let write_collection_begin (output: write_handler): unit = (
		write_value output VaCollection
	);;
	
	let write_collection_end = write_list_end;;
	
	let write_set_begin (output: write_handler): unit = (
		write_value output VaSet
	);;
	
	let write_set_end (output: write_handler): unit = (
		output "\x00" 0 1
	);;
	
	let write_set_element (output: write_handler)
		(x: [`int of int | `char of char | `wchar of int | `tag of string])
		: unit =
	(
		write_raw_short_string output (
			match x with
			| `int n -> string_of_int n
			| `char n -> "#" ^ string_of_int (int_of_char n)
			| `wchar n -> "#" ^ string_of_int n
			| `tag n -> n)
	);;

	let write_signature (output: write_handler): unit = (
		output filer_signature 0 (String.length filer_signature)
	);;
	
	let write_component (output: write_handler)
		(flags: filer_flags) (child_pos: int)
		(classname: string) (objectname: string): unit =
	(
		let write_prefix (output: write_handler)
			(flags: filer_flags) (child_pos: int): unit =
		(
			if flags <> 0 then (
				write_raw_byte (String.create 1) output (0xf0 lor flags);
				if flags land (set_of_ff FfChildPos) <> 0 then (
					write_int output child_pos (* not raw *)
				)
			)
		) in
		write_prefix output flags child_pos;
		write_raw_short_string output classname;
		write_raw_short_string output objectname
	);;
		
	let write_variant (output: write_handler) (value: variant): unit = (
		begin match value with
		| `bool x -> write_bool output x
		| `binary x -> write_binary output x
		| `char x -> write_char output x
		| `currency x -> write_currency output x
		| `efloat (_, image) -> write_raw_extended_image output image
		| `float x -> write_double output x
		| `int x -> write_int output x
		| `int32 x -> write_int32 output x
		| `int64 x -> write_int64 output x
		| `nil -> write_nil output
		| `null -> write_null output
		| `sfloat x -> write_single output x
		| `string x -> write_string output x
		| `tag x -> write_ident output x
		| `time x -> write_date output x
		| `wchar x -> write_wchar output x
		| `wstring x -> write_wide_string output x
		end
	)
	
	let create ?(resource_header: bool = true)
		(name: string) (out: out_channel)
		: [> `writing of (unit -> unit) list ref * (string -> 'v -> unit)] =
	(
		let output_channel = output out in
		let buf, output = (
			if resource_header then (
				let buf = Buffer.create 4096 in
				Some buf, Buffer.add_substring buf
			) else (
				None, output_channel
			)
		) in
		let state = ref [] in
		`writing (ref [], fun key value ->
			begin match value with
			| `end_of_record | `end_of_list ->
				begin match !state with
				| `component :: sr ->
					write_list_end output; (* properties *)
					write_list_end output; (* children *)
					state := sr
				| `persistent _ :: sr ->
					state := sr
				| `list1 :: sr ->
					write_list_begin output;
					write_list_end output;
					state := sr
				| (`list2 | `item | `collection) :: sr ->
					(* end of list equals end of item/end of collection *)
					write_list_end output;
					state := sr
				| `some :: _ | [] -> failwith "Writer.create"
				end;
				begin match buf with
				| Some buf when !state = [] ->
					write_resource_header output_channel name (Buffer.length buf);
					Buffer.output_buffer out buf;
					Buffer.clear buf
				| _ -> ()
				end
			| (`record | `list _ | `none | `some | #variant) as value ->
				begin match !state with
				| (`component | `persistent _ | `item) :: _ ->
					if key = "" then failwith "Writer.create:key was required"
				| (`list1 | `list2 | `collection | `some) :: _ | [] ->
					if key <> "" then failwith "Writer.create:key was not required"
				end;
				begin match value with
				| `record ->
					begin match !state with
					| (`component | `item) :: _ ->
						state := `persistent key :: !state
					| `persistent path :: _ ->
						state := `persistent (path ^ "." ^ key) :: !state
					| `list1 :: sr ->
						write_collection_begin output;
						write_list_begin output;
						state := `item :: `collection :: sr
					| `collection :: _ ->
						state := `item :: !state
					| [] ->
						write_signature output;
						write_component output 0 (-1) name "";
						state := `component :: []
					| `list2 :: _ | `some :: _ -> failwith "Writer.create"
					end
				| `list k ->
					begin match !state with
					| (`component | `item) :: _ ->
						write_property_name output key
					| `persistent path :: _ ->
						write_property_name output (path ^ "." ^ key)
					| `list1 :: sr ->
						write_list_begin output;
						state := `list2 :: sr
					| `list2 :: _ -> ()
					| `collection :: _ | `some :: _ | [] ->
						failwith "Writer.create"
					end;
					begin match k with
					| `any ->
						state := `list1 :: !state
					| `record ->
						write_collection_begin output;
						state := `collection :: !state
					end					
				| `none ->
					()
				| (#variant | `some) as value ->
					begin match !state with
					| (`component | `item) :: _ ->
						write_property_name output key
					| `persistent path :: _ ->
						write_property_name output (path ^ "." ^ key)
					| `list1 :: sr ->
						write_list_begin output;
						state := `list2 :: sr
					| `list2 :: _ -> ()
					| `some :: sr ->
						state := sr
					| `collection :: _ | [] -> failwith "Writer.create"
					end;
					begin match value with
					| `some ->
						state := `some :: !state
					| #variant as value ->
						write_variant output value
					end
				end
			end
		)
	);;

end;;

module Reader = struct
	
	type read_handler = string -> int -> int -> int;;
	
	let rec really (input: read_handler) s ofs len = (
		let r = input s ofs len in
		if r < len then (
			if len = 0 then raise End_of_file else
			really input s (ofs + r) (len - r)
		)
	);;
	
	let force (input: read_handler) length = (
		let s = String.create length in
		really input s 0 length;
		s
	);;
	
	let read_raw_byte (temp: string) (input: read_handler): int = (
		really input temp 0 1;
		int_of_char temp.[0]
	);;

	let read_raw_int8 (input: read_handler): int = (
		let s = force input 1 in
		let r = int_of_char s.[0] in
		if r >= 0x80 then r lor (lnot 0x7f) else r
	);;

	let read_raw_int16 (input: read_handler): int = (
		let s = force input 2 in
		let r = (int_of_char s.[0]) lor (int_of_char s.[1] lsl 8) in
		if r >= 0x8000 then r lor (lnot 0x7fff) else r
	);;
	
	let read_raw_int32 (input: read_handler): int32 = (
		let s = force input 4 in
		let r = Int32.of_int (int_of_char s.[0]) in
		let r = Int32.logor r (Int32.shift_left (Int32.of_int (int_of_char s.[1])) 8) in
		let r = Int32.logor r (Int32.shift_left (Int32.of_int (int_of_char s.[2])) 16) in
		let r = Int32.logor r (Int32.shift_left (Int32.of_int (int_of_char s.[3])) 24) in
		r
	);;
	
	let read_raw_int64 (input: read_handler): int64 = (
		let s = force input 8 in
		let r = Int64.of_int (int_of_char s.[0]) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[1])) 8) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[2])) 16) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[3])) 24) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[4])) 32) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[5])) 40) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[6])) 48) in
		let r = Int64.logor r (Int64.shift_left (Int64.of_int (int_of_char s.[7])) 56) in
		r
	);;
	
	let read_raw_extended (input: read_handler): float * (int64 * int) = (
		let m = read_raw_int64 input in
		let e = read_raw_int16 input in
		let pair = m, e in
		float_of_extended pair, pair
	);;
	
	let read_raw_date (input: read_handler): float = (
		unixtime_of_datetime (Int64.float_of_bits (read_raw_int64 input))
	);;
	
	let read_raw_short_string (input: read_handler): string = (
		let length = read_raw_byte (String.create 1) input in
		force input length
	);;

	let read_raw_zero_terminated_string (input: read_handler): string = (
		let rec loop s buf input = (
			really input s 0 1;
			let c = s.[0] in
			if c = '\x00' then Buffer.contents buf else (
				Buffer.add_char buf c;
				loop s buf input
			)
		) in
		loop (String.create 1) (Buffer.create 256) input
	);;

	let read_raw_long_string (input: read_handler): string = (
		let length = Int32.to_int (read_raw_int32 input) in
		force input length
	);;
	
	let read_raw_wide_string (input: read_handler): Unicode.utf16_string = (
		let length = Int32.to_int (read_raw_int32 input) in
		let result = Unicode.UTF16.create length in
		let block_length = 2048 in
		let block_size = block_length * 2 in
		let block = String.create block_size in
		let blocks = length / block_length in
		let rem = length mod block_length in
		for i = 0 to blocks - 1 do
			really input block 0 block_size;
			let offset = i * block_length in
			for j = 0 to block_length - 1 do
				let d = int_of_char block.[j * 2] lor
					(int_of_char block.[j * 2 + 1] lsl 8)
				in
				Bigarray.Array1.set result (offset + j) d
			done
		done;
		really input block 0 (rem * 2);
		let offset = blocks * block_length in
		for j = 0 to rem - 1 do
			let d = int_of_char block.[j * 2] lor
				(int_of_char block.[j * 2 + 1] lsl 8)
			in
			Bigarray.Array1.set result (offset + j) d
		done;
		result
	);;
	
	let read_raw_binary (input: read_handler): byte_array = (
		let length = Int32.to_int (read_raw_int32 input) in
		let result = Bigarray.Array1.create
			Bigarray.int8_unsigned Bigarray.c_layout length
		in
		let block_size = 4096 in
		let block = String.create block_size in
		let blocks = length / block_size in
		let rem = length mod block_size in
		for i = 0 to blocks - 1 do
			really input block 0 block_size;
			let offset = i * block_size in
			for j = 0 to block_size - 1 do
				Bigarray.Array1.set result (offset + j) (int_of_char block.[j])
			done
		done;
		really input block 0 rem;
		let offset = blocks * block_size in
		for j = 0 to rem - 1 do
			Bigarray.Array1.set result (offset + j) (int_of_char block.[j])
		done;
		result
	);;
	
	let read_value (input: read_handler): value_type = (
		let result = Obj.magic (read_raw_byte (String.create 1) input: int) in
		if (result < min_value || result > max_value) then (
			failwith "read_value"
		);
		result
	);;
	
	let read_int (input: read_handler): int64 = (
		let v = read_value input in
		match v with
		| VaInt8 -> Int64.of_int (read_raw_int8 input)
		| VaInt16 -> Int64.of_int (read_raw_int16 input)
		| VaInt32 -> Int64.of_int32 (read_raw_int32 input)
		| VaInt64 -> read_raw_int64 input
		| _ -> failwith "Reader.read_int"
	);;
	
	let read_set_element (input: read_handler)
		: [`int of int | `char of char | `wchar of int | `tag of string] option =
	(
		let len = read_raw_byte (String.create 1) input in
		if len = 0 then None else
		let r = force input len in
		Some (
			if r.[0] = '#' then (
				let c = int_of_string (String.sub r 1 (len - 1)) in
				if c <= 0x7f then `char (char_of_int c) else `wchar c
			) else (
				try `int (int_of_string r) with
				| Failure _ -> `tag r
			)
		)
	);;
	
	let read_list (input: read_handler): int option = (
		begin match read_value input with
		| VaNull -> None
		| VaList -> Some (-1)
		| VaInt8 -> Some (read_raw_int8 input)
		| VaInt16 -> Some (read_raw_int16 input)
		| VaInt32 -> Some (Int32.to_int (read_raw_int32 input))
		| VaInt64 -> Some (Int64.to_int (read_raw_int64 input))
		| _ -> failwith "read_list"
		end
	);;
		
	let read_variant (input: read_handler)
		: [`list | `set | `collection | variant] =
	(
		begin match read_value input with
		| VaNull -> `null
		| VaList -> `list
		| VaInt8 -> `int (read_raw_int8 input)
		| VaInt16 -> `int (read_raw_int16 input)
		| VaInt32 -> `int32 (read_raw_int32 input)
		| VaExtended -> `efloat (read_raw_extended input)
		| VaString -> `string (read_raw_short_string input)
		| VaIdent -> `tag (read_raw_short_string input)
		| VaFalse -> `bool false
		| VaTrue -> `bool true
		| VaBinary -> `binary (read_raw_binary input)
		| VaSet -> `set
		| VaLString -> `string (read_raw_long_string input)
		| VaNil -> `nil
		| VaCollection -> `collection
		| VaSingle -> `sfloat (Int32.float_of_bits (read_raw_int32 input))
		| VaCurrency -> `currency (Int64.to_float (read_raw_int64 input) /. 10000.0)
		| VaDate -> `time (read_raw_date input)
		| VaWString -> `wstring (read_raw_wide_string input)
		| VaInt64 -> `int64 (read_raw_int64 input)
		| VaUTF8String -> `string (read_raw_long_string input)
		| VaDouble -> `float (Int64.float_of_bits (read_raw_int64 input))
		end		
	);;
	
	let read_property_name (input: read_handler): string option = (
		let r = read_raw_byte (String.create 1) input in
		if r = 0 then None else Some (force input r)
	);;
	
	let read_resource_header (input: read_handler): string = (
		let temp = String.create 4 in
		really input temp 0 3;
		if temp.[0] <> '\xff' &&
			temp.[1] <> '\x0a' &&
			temp.[2] <> '\x00' then failwith "read_resouce_header.1" else
		let name = read_raw_zero_terminated_string input in
		really input temp 0 2;
		if temp.[0] <> '\x30' &&
			temp.[1] <> '\x10' then failwith "read_resouce_header.2" else
		really input temp 0 4; (* size *)
		name
	);;

	let read_signature (input: read_handler): unit = (
		let signature = force input (String.length filer_signature) in
		if signature <> filer_signature then failwith "read_signature"
	);;

	let read_component (input: read_handler)
		: (filer_flags * int * string * string) option =
	(
		let first = read_raw_byte (String.create 1) input in
		if first = 0 then (
			None
		) else if first land 0xf0 = 0xf0 then (
			let flags = first land 0x0f in
			let child_pos = (
				if flags land (set_of_ff FfChildPos) <> 0 then (
					Int64.to_int (read_int input) (* not raw *)
				) else (
					-1
				)
			) in
			let classname = read_raw_short_string input in
			let objectname = read_raw_short_string input in
			Some (flags, child_pos, classname, objectname)
		) else (
			let classname = force input first in
			let objectname = read_raw_short_string input in
			Some (0, -1, classname, objectname)
		)
	);;

	let create ?(resource_header: bool = true)
		(name: string) (in_file: in_channel)
		: [> `reading of string ref * 'v ref * (unit -> unit) *
			(float -> float)] =
	(
		let input = input in_file in
		let state = ref [] in
		let next_key = ref "" in
		let next_value = ref `end_of_record in
		let advance (): unit = (
			let convert_value value = (
				begin match value with
				| `binary x ->
					next_value := `binary x
				| `bool x ->
					next_value := `bool x
				| `char x ->
					next_value := `char x
				| `wchar x ->
					next_value := `char (char_of_int x)
				| `float x | `sfloat x | `efloat (x, _) | `currency x ->
					next_value := `float x
				| `int x ->
					next_value := `int x
				| `int32 x ->
					next_value := (
						let i = Int32.to_int x in
						if Int32.of_int i = x then `int i else
						`int32 x
					)
				| `int64 x ->
					next_value := (
						let i = Int64.to_int x in
						if Int64.of_int i = x then `int i else
						let i = Int64.to_int32 x in
						if Int64.of_int32 i = x then `int32 i else
						`int64 x
					)
				| `nil ->
					next_value := `nil
				| `null ->
					next_value := `null
				| `string x ->
					next_value := (
						if String.length x = 1 then `char x.[0] else
						`string x
					)
				| `wstring x ->
					next_value := (
						let u = Unicode.utf8_of_utf16 x in
						if String.length u = 1 then `char u.[0] else
						`string u
					)
				| `tag x ->
					next_value := `tag x
				| `time x ->
					next_value := `time x
				| (`collection | `list | `set) as g ->
					next_value := `list `any;
					state := g :: !state
				end
			) in
			let leave_component sr = (
				if read_component input <> None then (
					failwith "Reader.create"
				);
				next_key := "";
				next_value := `end_of_record;
				state := sr
			) in
			let property_name_down parent path sr = (
				if String.contains path '.' then (
					let p = String.index path '.' in
					next_key := String.sub path 0 p;
					next_value := `record;
					state := `property_name_down ((
						if parent <> "" then parent ^ "." ^ !next_key else
						!next_key),
						let n = p + 1 in
						String.sub path n (String.length path - n)) :: sr
				) else (
					next_key := path;
					state := sr;
					convert_value (read_variant input)
				)
			) in
			let property_name_up parent path sr = (
				let parent_length = String.length parent in
				let length = String.length path in
				if length > parent_length && path.[parent_length] = '.' &&
					String.sub path 0 parent_length = parent
				then (
					let n = parent_length + 1 in
					let diff_name = String.sub path n (length - n) in
					property_name_down parent diff_name sr
				) else (
					next_key := "";
					next_value := `end_of_record;
					assert (match List.hd sr with
						| `persistent _ -> true
						| _ -> false);
					state := `property_name_up path :: List.tl sr
				)
			) in
			begin match !state with
			| `component :: sr ->
				begin match read_property_name input with
				| Some x ->
					property_name_down "" x !state
				| None ->
					leave_component sr
				end
			| `item :: sr ->
				begin match read_property_name input with
				| Some x ->
					property_name_down "" x !state
				| None ->
					next_key := "";
					next_value := `end_of_record;
					state := sr
				end
			| `persistent parent :: sr ->
				begin match read_property_name input with
				| Some x ->
					property_name_up parent x !state
				| None ->
					next_key := "";
					next_value := `end_of_record;
					state := `end_properties :: sr
				end
			| `property_name_down (parent, path) :: sr ->
				property_name_down parent path (`persistent parent :: sr)
			| `property_name_up x :: ((`persistent parent :: _) as sr) ->
				property_name_up parent x sr
			| `property_name_up x :: sr ->
				property_name_down "" x sr
			| `end_properties :: `component :: sr ->
				leave_component sr
			| `end_properties :: `item :: sr ->
				next_key := "";
				next_value := `end_of_record;
				state := sr
			| `end_properties :: `persistent _ :: sr ->
				next_key := "";
				next_value := `end_of_record;
				state := `end_properties :: sr
			| `end_properties :: _ ->
				assert false (* invalid state *)
			| `collection :: sr ->
				begin match read_list input with
				| Some _ ->
					next_key := "";
					next_value := `record;
					state := `item :: !state
				| None ->
					next_key := "";
					next_value := `end_of_list;
					state := sr
				end
			| `list :: sr ->
				begin match read_variant input with
				| `null ->
					next_key := "";
					next_value := `end_of_list;
					state := sr
				| x ->
					next_key := "";
					convert_value x
				end
			| `set :: sr ->
				begin match read_set_element input with
				| Some x ->
					next_key := "";
					convert_value x
				| None ->
					next_key := "";
					next_value := `end_of_list;
					state := sr
				end
			| [] ->
				if resource_header then (
					let (_: string) = read_resource_header input in ()
				);
				read_signature input;
				begin match read_component input with
				| Some (_, _, classname, _) ->
					if classname <> name then (
						failwith ("Reader.create:mismatch \'" ^ classname ^ "\'")
					);
					next_key := "";
					next_value := `record;
					state := `component :: []
				| None -> failwith "Reader.create"
				end
			end
		) in
		advance ();
		`reading (next_key, next_value, advance, unixtime_of_datetime)
	);;

end;;
